#!/usr/bin/env python3

"""
#define  FT_FILENAME			 0x01	// <string>
#define  FT_FILESIZE			 0x02	// <uint32> (or <uint64> when supported)
#define  FT_FILESIZE_HI			 0x3A	// <uint32>
#define  FT_FILETYPE			 0x03	// <string>
#define  FT_FILEFORMAT			 0x04	// <string>
#define  FT_LASTSEENCOMPLETE	 0x05	// <uint32>
#define	TAG_PART_PATH			"\x06"	// <string>
#define	TAG_PART_HASH			"\x07"
#define  FT_TRANSFERRED			 0x08	// <uint32>
#define	TAG_TRANSFERRED			"\x08"	// <uint32>
#define  FT_GAPSTART			 0x09	// <uint32>
#define	TAG_GAPSTART			"\x09"	// <uint32>
#define  FT_GAPEND				 0x0A	// <uint32>
#define	TAG_GAPEND				"\x0A"	// <uint32>
#define  FT_DESCRIPTION			 0x0B	// <string>
#define	TAG_DESCRIPTION			"\x0B"	// <string>
#define	TAG_PING				"\x0C"
#define	TAG_FAIL				"\x0D"
#define	TAG_PREFERENCE			"\x0E"
#define  FT_PARTFILENAME		 0x12	// <string>
//#define FT_PRIORITY			 0x13	// Not used anymore
#define  FT_STATUS				 0x14	// <uint32>
#define  FT_SOURCES				 0x15	// <uint32>
#define  FT_PERMISSIONS			 0x16	// <uint32>
//#define FT_ULPRIORITY			 0x17	// Not used anymore
#define  FT_DLPRIORITY			 0x18	// Was 13
#define  FT_ULPRIORITY			 0x19	// Was 17
#define	 FT_COMPRESSION			 0x1A
#define	 FT_CORRUPTED			 0x1B
#define  FT_KADLASTPUBLISHKEY	 0x20	// <uint32>
#define  FT_KADLASTPUBLISHSRC	 0x21	// <uint32>
#define	 FT_FLAGS				 0x22	// <uint32>
#define	 FT_DL_ACTIVE_TIME		 0x23	// <uint32>
#define	 FT_CORRUPTEDPARTS		 0x24	// <string>
#define  FT_DL_PREVIEW			 0x25
#define  FT_KADLASTPUBLISHNOTES	 0x26	// <uint32>
#define  FT_AICH_HASH			 0x27
#define  FT_FILEHASH			 0x28
#define	 FT_COMPLETE_SOURCES	 0x30	// nr. of sources which share a complete version of the associated file (supported by eserver 16.46+)
#define  FT_COLLECTIONAUTHOR	 0x31
#define  FT_COLLECTIONAUTHORKEY  0x32
#define  FT_PUBLISHINFO			 0x33	// <uint32>
#define  FT_LASTSHARED			 0x34	// <uint32>
#define  FT_AICHHASHSET			 0x35	// <uint32>
#define	TAG_KADAICHHASHPUB		"\x36"	// <AICH Hash>

// statistic
#define  FT_ATTRANSFERRED		 0x50	// <uint32>
#define  FT_ATREQUESTED			 0x51	// <uint32>
#define  FT_ATACCEPTED			 0x52	// <uint32>
#define  FT_CATEGORY			 0x53	// <uint32>
#define	 FT_ATTRANSFERREDHI		 0x54	// <uint32>
#define	 FT_MAXSOURCES			 0x55	// <uint32>
#define	 FT_MEDIA_ARTIST		 0xD0	// <string>
#define	TAG_MEDIA_ARTIST		"\xD0"	// <string>
#define	 FT_MEDIA_ALBUM			 0xD1	// <string>
#define	TAG_MEDIA_ALBUM			"\xD1"	// <string>
#define	 FT_MEDIA_TITLE			 0xD2	// <string>
#define	TAG_MEDIA_TITLE			"\xD2"	// <string>
#define	 FT_MEDIA_LENGTH		 0xD3	// <uint32> !!!
#define	TAG_MEDIA_LENGTH		"\xD3"	// <uint32> !!!
#define	 FT_MEDIA_BITRATE		 0xD4	// <uint32>
#define	TAG_MEDIA_BITRATE		"\xD4"	// <uint32>
#define	 FT_MEDIA_CODEC			 0xD5	// <string>
#define	TAG_MEDIA_CODEC			"\xD5"	// <string>
#define  FT_FILECOMMENT			 0xF6	// <string>
#define  FT_FILERATING			 0xF7	// <uint8>
"""

import os
import binascii
from enum import Enum
from collections import namedtuple


# METENTRY_FIELDS = ('FT_FILENAME', 'ED2KHASH', 'NUMPARTS', 'LASTCHANGED', 'PARTHASHES', 'FT_FILESIZE',
#                    'FT_ATTRANSFERRED', 'FT_ATREQUESTED', 'FT_ATACCEPTED', 'FT_ATTRANSFERREDHI', 'FT_PARTFILENAME',
#                    'FT_AICHHASHSET', 'FT_AICH_HASH', 'FT_ULPRIORITY', 'FT_KADLASTPUBLISHSRC')
# MetEntry = namedtuple('met_entry', METENTRY_FIELDS)

HEADER_KNOWN_MET_FILE = b'\x0E'
HEADER_PART_MET_FILE = b'\xE0'


# https://github.com/amule-project/amule/blob/master/src/include/tags/TagTypes.h
# https://github.com/amule-project/amule/blob/master/src/include/tags/FileTags.h


# TYPE: \x02
class TagString(Enum):
    FT_FILENAME = 0x01
    FT_FILETYPE = 0x03
    FT_FILEFORMAT = 0x04
    FT_PARTFILENAME = 0x12
    FT_AICH_HASH = 0x27
    FT_CORRUPTEDPARTS = 0x24
    FT_MEDIA_ARTIST = 0xD0
    FT_MEDIA_ALBUM = 0xD1
    FT_MEDIA_TITLE = 0xD2
    FT_MEDIA_CODEC = 0xD5


# TYPE: \x03
class TagUint32(Enum):
    FT_FILESIZE = 0x02
    FT_FILESIZE_HI = 0x3A
    FT_LASTSEENCOMPLETE = 0x05
    FT_TRANSFERRED = 0x08
    FT_GAPSTART = 0x09
    FT_GAPEND = 0x0A
    FT_STATUS = 0x14
    FT_SOURCES = 0x15
    FT_PERMISSIONS = 0x16
    FT_DLPRIORITY = 0x18
    FT_ULPRIORITY = 0x19
    FT_KADLASTPUBLISHKEY = 0x20
    FT_KADLASTPUBLISHSRC = 0x21
    FT_FLAGS = 0x22
    FT_DL_ACTTVE_TIME = 0x23
    FT_COMPLETE_SOURCES = 0x30
    FT_PUBLISHINFO = 0x33
    FT_LASTSHARED = 0x34
    FT_ATTRANSFERRED = 0x50
    FT_ATTRANSFERREDHI = 0x54
    FT_ATREQUESTED = 0x51
    FT_ATACCEPTED = 0x52
    FT_CATEGORY = 0x53
    FT_MEDIA_LENGTH = 0xD3
    FT_MEDIA_BITRATE = 0xD4


# TYPE: \x09
class TagUint8(Enum):
    FT_FILERATING = 0xF7


# TYPE: \x07
class TagBlob(Enum):
    FT_AICHHASHSET = 0x35


class TagDeprecated(Enum):
    FT_OLDDLPRIORITY = 0x13
    FT_OLDULPRIORITY = 0x17


class TagHash(Enum):
    NA = 0x00


TAG_TYPE = {
    b'\x01': None,  # HASH
    b'\x02': TagString,
    b'\x03': TagUint32,
    b'\x04': None,  # FLOAT32
    b'\x05': None,  # BOOL
    b'\x06': None,  # BOOLARRAY
    b'\x07': TagBlob,
    b'\x08': None,  # UINT16
    b'\x09': TagUint8,
    b'\x0A': None,  # BSOB
    b'\x0B': None,  # UINT64

    b'\x12': None,  # STRnn
    b'\x13': None,  # STRnn
    b'\x14': None,  # STRnn
    b'\x15': None,  # STRnn
    b'\x16': None,  # STRnn
    b'\x17': None,  # STRnn
    b'\x18': None,  # STRnn
    b'\x19': None,  # STRnn
    b'\x1A': None,  # STRnn
    b'\x11': None,  # STRnn
    b'\x1B': None,  # STRnn
    b'\x1C': None,  # STRnn
    b'\x1D': None,  # STRnn
    b'\x1E': None,  # STRnn
    b'\x1F': None,  # STRnn
    b'\x20': None,  # STRnn
    b'\x21': None,  # STRnn
    b'\x22': None,  # STRnn
    b'\x23': None,  # STRnn
    b'\x24': None,  # STRnn
    b'\x25': None,  # STRnn
    b'\x26': None,  # STRnn
}

MetEntry = namedtuple('met_entry',
                      ['ED2KHASH', 'NUMPARTS', 'LASTCHANGED', 'PARTHASHES','TAGCOUNT']
                      + [x.name for x in TagString]
                      + [x.name for x in TagUint32]
                      + [x.name for x in TagBlob])
MetEntry.__new__.__defaults__ = (None,) * len(MetEntry._fields)


def met_read_string(fd):
    string_len = int.from_bytes(fd.read(2), byteorder='little')
    return fd.read(string_len).decode('utf-8', errors='replace')


def met_read_integer(fd, size=4):
    return int.from_bytes(fd.read(size), byteorder='little')


def met_read_blob(fd):
    blob_size = met_read_integer(fd)
    return fd.read(blob_size)


def record_last_changed(fd):
    return met_read_integer(fd)


def record_read_hash16(fd):
    aichash = fd.read(16)
    return binascii.hexlify(aichash).decode('utf-8')


def record_read_tags(fd):
    tags = dict()
    tags['TAGCOUNT'] = met_read_integer(fd)
    for index in range(tags['TAGCOUNT']):
        tag_buf = fd.read(4)
        tag_type = bytes([tag_buf[0]])
        tag_id = tag_buf[-1]

        print('Reading at 0x{:08X}: 0x{}'.format(fd.tell()-4, tag_buf.hex()))

        # Validate TAG bytes
        if tag_type not in TAG_TYPE and not tag_buf[1:3] == b'\x01\x00':
            print('<<< Premature end of record at offset 0x{:08X} ({})'.format(fd.tell(), tag_buf))
            return tags

        if TAG_TYPE[tag_type] is None:
            print('... Unsupported {} with tag: {}'.format(tag_type, tag_id))
            continue

        tag_enum = TAG_TYPE[tag_type]
        if tag_enum is TagString:
            tags[tag_enum(tag_id).name] = met_read_string(fd)
        elif tag_enum is TagBlob:
            tags[tag_enum(tag_id).name] = met_read_blob(fd)
        elif tag_enum is TagUint32:
            if tag_id in [0x09, 0x0A]:
                fd.seek(5, 1)
                continue
            tags[tag_enum(tag_id).name] = met_read_integer(fd)
        elif tag_enum is TagUint8:
            tags[tag_enum(tag_id).name] = met_read_integer(fd, size=1)

    return tags


def record_num_parts(fd):
    return met_read_integer(fd, 2)


def load_record(fd):
    met_tags = {}

    met_tags['LASTCHANGED'] = record_last_changed(fd)
    met_tags['ED2KHASH'] = record_read_hash16(fd)
    met_tags['NUMPARTS'] = record_num_parts(fd)
    partshashes = []
    for i in range(met_tags['NUMPARTS']):
        partshashes.append(record_read_hash16(fd))
    met_tags['PARTHASHES'] = partshashes
    met_tags.update(record_read_tags(fd))

    return MetEntry(**met_tags)


def load_knownfiles(filename):
    with open(filename, 'rb') as fd:
        magic = fd.read(1)
        if os.path.basename(filename) == 'known.met':
            count = int.from_bytes(fd.read(4), byteorder='little')
            # print('Number of records:   {}'.format(count))
        if magic == HEADER_KNOWN_MET_FILE:
            for index in range(count):
                # print('>>> [{}] NEW RECORD ***'.format(index+1))
                yield load_record(fd)
        elif magic == HEADER_PART_MET_FILE:
            # print('>>> [{}] NEW RECORD ***'.format(0))
            yield load_record(fd)